/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.plugins.checkstyle;

import ch.hortis.sonar.model.ActiveRule;
import ch.hortis.sonar.model.Rule;
import ch.hortis.sonar.model.RuleFailureLevel;
import ch.hortis.sonar.model.RulesProfile;
import com.thoughtworks.xstream.XStream;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.sonar.commons.rules.ActiveRuleParam;
import org.sonar.commons.rules.RuleParam;
import org.sonar.plugins.api.Language;
import org.sonar.plugins.api.Languages;
import org.sonar.plugins.api.rules.ConfigurationExportable;
import org.sonar.plugins.api.rules.ConfigurationImportable;
import org.sonar.plugins.api.rules.RulesRepository;
import org.sonar.plugins.api.rules.StandardRulesXmlParser;
import org.sonar.plugins.checkstyle.xml.Module;
import org.sonar.plugins.checkstyle.xml.Property;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class CheckstyleRulesRepository implements RulesRepository, ConfigurationExportable, ConfigurationImportable {
  private static final String MODULE_SEPARATOR = "/";

  public Language getLanguage() {
    return Languages.JAVA;
  }

  public List<Rule> getInitialReferential() {
    InputStream input = getClass().getResourceAsStream("/org/sonar/plugins/checkstyle/rules.xml");
    try {
      return new StandardRulesXmlParser().parse(input);

    } finally {
      IOUtils.closeQuietly(input);
    }
  }

  public List<Rule> parseReferential(String fileContent) {
    return new StandardRulesXmlParser().parse(fileContent);
  }

  public List<RulesProfile> getProvidedProfiles() {
    return Arrays.asList(
      loadProvidedProfile(RulesProfile.SONAR_WAY_NAME, "/org/sonar/plugins/checkstyle/profile-sonar-way.xml"),
      loadProvidedProfile(RulesProfile.SUN_CONVENTIONS_NAME, "/org/sonar/plugins/checkstyle/profile-sun-conventions.xml"));
  }

  private RulesProfile loadProvidedProfile(String name, String filename) {
    try {
      InputStream sonarWayProfile = getClass().getResourceAsStream(filename);
      RulesProfile profile = new RulesProfile(name);

      // TODO set as mandatory into the RulesProfile constructor
      profile.setLanguage(Languages.JAVA.getKey());
      profile.setActiveRules(importConfiguration(IOUtils.toString(sonarWayProfile), getInitialReferential()));
      return profile;

    } catch (IOException e) {
      throw new RuntimeException("Configuration file not found for the profile : " + name, e);
    }
  }

  public String exportConfiguration(RulesProfile activeProfile) {
    Module tree = buildModuleTree(activeProfile.getActiveRulesByPlugin(CheckstylePlugin.KEY));
    String xmlModules = buildXmlFromModuleTree(tree);
    return addHeaderToXml(xmlModules);
  }

  public List<ActiveRule> importConfiguration(String configuration, List<Rule> rules) {
    List<ActiveRule> activeRules = new ArrayList<ActiveRule>();
    Module moduleTree = buildModuleTreeFromXml(configuration);
    buildActiveRulesFromModuleTree(moduleTree, "", activeRules, rules);
    return activeRules;
  }

  protected Module buildModuleTree(List<ActiveRule> activeRules) {
    Module root = new Module("");
    for (ActiveRule activeRule : activeRules) {
      if (activeRule.getRule().getPluginName().equals(CheckstylePlugin.KEY)) {
        String configKey = activeRule.getRule().getConfigKey();
        Module current = root;
        String[] modules = StringUtils.split(configKey, MODULE_SEPARATOR);
        for (int modulePos = 0; modulePos < modules.length; modulePos++) {
          String moduleName = modules[modulePos];
          if (current.getChildren().isEmpty()) {
            Module child = new Module(moduleName);
            current.addChild(child);
            current = child;
          } else {
            int index = current.getChildPosition(moduleName);
            // Create new module if it doesn't exists or if it's the last one (To allow same rule activation)
            if (index < 0 || modulePos == modules.length - 1) {
              Module child = new Module(moduleName);
              current.addChild(child);
              current = child;
            } else {
              current = current.getChildren().get(index);
            }
          }
        }
        List<Property> properties = new ArrayList<Property>();
        String failureLevel = StringUtils.lowerCase(activeRule.getLevel().name());
        properties.add(new Property("severity", failureLevel));
        for (ActiveRuleParam activeRuleParam : activeRule.getActiveRuleParams()) {
          if (activeRuleParam.getValue() != null && activeRuleParam.getValue().length() != 0) {
            properties.add(new Property(activeRuleParam.getRuleParam().getKey(), activeRuleParam.getValue()));
          }
        }
        current.setProperties(properties);
      }
    }
    if (root.getChildren().isEmpty()) {
      return new Module("Checker", null);
    } else {
      return root.getChildren().get(0);
    }
  }

  protected String buildXmlFromModuleTree(Module tree) {
    XStream xstream = new XStream();
    xstream.processAnnotations(Module.class);
    xstream.processAnnotations(Property.class);
    return xstream.toXML(tree);
  }

  protected String addHeaderToXml(String xmlModules) {
    String header = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
      "<!DOCTYPE module PUBLIC \"-//Puppy Crawl//DTD Check Configuration 1.2//EN\" \"http://www.puppycrawl.com/dtds/configuration_1_2.dtd\">" +
      "<!-- generated by Sonar -->\n";
    return header + xmlModules;
  }

  protected Module buildModuleTreeFromXml(String configuration) {
    XStream xStream = new XStream();
    xStream.processAnnotations(Module.class);
    xStream.processAnnotations(Property.class);
    return (Module) xStream.fromXML(configuration);
  }

  protected void buildActiveRulesFromModuleTree(Module moduleTree, String modulePath, List<ActiveRule> activeRules, List<Rule> rules) {
    if (moduleTree.getChildren() == null || moduleTree.getChildren().isEmpty()) {
      for (Rule rule : rules) {
        if (rule.getConfigKey().equals(modulePath)) {
          RuleFailureLevel ruleFailureLevel = getRuleFailureLevel(moduleTree);
          if (ruleFailureLevel != null) {
            ActiveRule activeRule = new ActiveRule(null, rule, ruleFailureLevel);
            activeRule.setActiveRuleParams(getActiveRuleParams(moduleTree, rule, activeRule));
            activeRules.add(activeRule);
            return;
          }
        }
      }
    } else {
      String baseModulePath = modulePath.length() == 0 ? moduleTree.getName() + MODULE_SEPARATOR :
        modulePath + MODULE_SEPARATOR;
      for (Module child : moduleTree.getChildren()) {
        buildActiveRulesFromModuleTree(child, baseModulePath + child.getName(), activeRules, rules);
      }
    }
  }

  private RuleFailureLevel getRuleFailureLevel(Module module) {
    if (module.getProperties() != null) {
      for (Property property : module.getProperties()) {
        if (property.getName().equals("severity")) {
          if (property.getValue().equals(StringUtils.lowerCase(RuleFailureLevel.ERROR.toString()))) {
            return RuleFailureLevel.ERROR;

          } else if (property.getValue().equals(StringUtils.lowerCase(RuleFailureLevel.WARNING.toString()))) {
            return RuleFailureLevel.WARNING;

          } else {
            return null;
          }
        }
      }
    }
    return RuleFailureLevel.ERROR;
  }

  private List<ActiveRuleParam> getActiveRuleParams(Module module, Rule rule, ActiveRule activeRule) {
    List<ActiveRuleParam> activeRuleParams = new ArrayList<ActiveRuleParam>();
    if (module.getProperties() != null) {
      for (Property property : module.getProperties()) {
        if (rule.getParams() != null) {
          for (RuleParam ruleParam : rule.getParams()) {
            if (ruleParam.getKey().equals(property.getName())) {
              activeRuleParams.add(new ActiveRuleParam(activeRule, ruleParam, property.getValue()));
            }
          }

        }
      }
    }
    return activeRuleParams;
  }

}
